"""     
###################################################################################
#                                                                                 #
#   Copyright "2015" "NEXT ENERGY"                                                #
#                                                                                 #
#   Licensed under the Apache License, Version 2.0 (the "License");               #
#   you may not use this file except in compliance with the License.              #
#   You may obtain a copy of the License at                                       #
#                                                                                 #
#       http://www.apache.org/licenses/LICENSE-2.0                                #
#                                                                                 #
#   Unless required by applicable law or agreed to in writing, software           #
#   distributed under the License is distributed on an "AS IS" BASIS,             #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.      #
#   See the License for the specific language governing permissions and           #
#   limitations under the License.                                                #
#                                                                                 #
###################################################################################
"""
import os
import logging
logging = logging.getLogger(os.path.basename(__file__))
from db_common import dbconn_from_args
import re
    

def abstract_2subs(cur,conn):
    """
    Abstracting relations which have exactly 2 stations to vertices and links.
    Vertices: Substations are in general polygons (which applies to the tags "substation", "generator", "station", "switch", 
    etc. The substations are abstracted as a vertex, which is the geometric center of their constituing polygon. 
    If a substation is a node in OpenStreetmap, then the vertex is this node.
    If a station is of type "relation" in OpenStreetMap, the rectangle which includes all nodes of the relation is 
    defined and its center is used as a vertex. 
    To define which relations in OpenStreetMap have exacly two substations, relations analysis is performed in
    relation_analysis.py and the results are to be found in the table _analysis_rels.
    
    Links: Relation are made of one or more links (transmission lines). These transmission lines are abstracted as the
    links between the two vertices obtained when abstracting the substations. The actual length of the transmission lines
    is calculated as sum of the line (and cables) segments length, which are also listed as constituing the relation.    
    """    
    
    # Create the abstraction function for power relations with two substations
    sql =   """
            CREATE OR REPLACE FUNCTION abstract_rel_with_2subs(rel_id bigint) 
            RETURNS text
            AS $$
            DECLARE 
                vert1_id                 bigint;
                vert2_id                 bigint;
                vert1_id_typ             char;
                vert2_id_typ             char;
                vert1_center             geometry;
                vert2_center             geometry;
                vert1_power              text;
                vert2_power              text;
                vert1_volt               text;
                vert2_volt               text;
                vert1_freq               text;
                vert2_freq               text;
                vert1_name               text;
                vert2_name               text;
                vert1_operator           text;
                vert2_operator           text;
                vert1_ref                text;
                vert2_ref                text;
                link_length              integer;
                link_geom                geometry;
                link_volt                integer;
                link_cables              smallint; 
                link_wires               text; 
                link_frequency           text;
                link_operator            text;
                link_name                text;
                link_ref                 text;
                link_transmissions       bigint[];
                link_id_1                bigint;
                link_id_2                bigint;
                link_wires_nb            integer;
            BEGIN
                vert1_id = (SELECT osm_id FROM get_rel_parts_with_power_vals(rel_id,'{station,substation,sub_station,plant,generator}'::text[]) ORDER BY osm_id ASC LIMIT 1);
                vert2_id = (SELECT osm_id FROM get_rel_parts_with_power_vals(rel_id,'{station,substation,sub_station,plant,generator}'::text[]) ORDER BY osm_id ASC LIMIT 1 OFFSET 1);
                vert1_id_typ = (SELECT osm_id_typ FROM get_rel_parts_with_power_vals(rel_id,'{station,substation,sub_station,plant,generator}'::text[]) ORDER BY osm_id ASC LIMIT 1);
                vert2_id_typ = (SELECT osm_id_typ FROM get_rel_parts_with_power_vals(rel_id,'{station,substation,sub_station,plant,generator}'::text[]) ORDER BY osm_id ASC LIMIT 1 OFFSET 1);
                IF vert1_id_typ = 'n' THEN 
                    vert1_center = (SELECT ST_SetSRID(ST_MakePoint(lon/100.0,lat/100.0),900913) FROM planet_osm_nodes WHERE id = vert1_id);
                    vert1_name = (SELECT hstore(tags)->'name' FROM planet_osm_nodes WHERE id=vert1_id );
                    vert1_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_nodes WHERE id=vert1_id );
                    vert1_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_nodes WHERE id=vert1_id );
                    vert1_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_nodes WHERE id = vert1_id);
                    vert1_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_nodes WHERE id = vert1_id);
                    vert1_power = (SELECT value FROM vertices_ref_names JOIN planet_osm_nodes ON  hstore(tags)->'power' = vertices_ref_names.name WHERE id = vert1_id);
                ELSE IF vert1_id_typ = 'w' THEN
                        vert1_center = (SELECT ST_centroid(way) FROM planet_osm_polygon WHERE osm_id = vert1_id);
                        vert1_name = (SELECT hstore(tags)->'name' FROM planet_osm_ways WHERE id=vert1_id );
                        vert1_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_ways WHERE id=vert1_id );
                        vert1_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_ways WHERE id=vert1_id );
                        vert1_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_ways WHERE id = vert1_id);
                        vert1_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_ways WHERE id = vert1_id);
                        vert1_power = (SELECT value FROM vertices_ref_names JOIN planet_osm_ways ON  hstore(tags)->'power' = vertices_ref_names.name WHERE id = vert1_id);
                    ELSE IF vert1_id_typ = 'r' THEN
                        vert1_center = (SELECT ST_SetSRID(ST_MakePoint((max(lon) + min(lon))/200.0,(max(lat) + min(lat))/200.0),900913) 
                                           FROM planet_osm_nodes WHERE id IN (
                                           SELECT trim(leading 'n' from member)::bigint as node_parts 
                                           FROM (
                                               SELECT unnest(members) as member,* 
                                               FROM planet_osm_rels WHERE id = vert1_id) t 
                                               WHERE member~E'[n]\\\\d+'
                                               UNION
                                               SELECT unnest(nodes) as node_parts 
                                               FROM planet_osm_ways 
                                               WHERE id IN (
                                                   SELECT trim(leading 'w' from member)::bigint as way_parts 
                                                   FROM (
                                                       SELECT unnest(members) as member 
                                                       FROM planet_osm_rels 
                                                       WHERE id = vert1_id) t 
                                                   WHERE member~E'[w]\\\\d+')));
                        vert1_name = (SELECT hstore(tags)->'name' FROM planet_osm_rels WHERE id=vert1_id );
                        vert1_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_rels WHERE id=vert1_id );
                        vert1_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_rels WHERE id=vert1_id );
                        vert1_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_rels WHERE id = vert1_id);
                        vert1_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_rels WHERE id = vert1_id);
                        vert1_power = (SELECT value FROM vertices_ref_names JOIN planet_osm_rels ON  hstore(tags)->'power' = vertices_ref_names.name WHERE id = vert1_id);
                        ELSE 
                            INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_2subs','ERROR', 'Cannot get vert1_center,vert1_name,vert1_operator,vert1_ref,vert1_volt,vert1_power');
                            RETURN 'ERROR: Cannot get vert1_center,vert1_name,vert1_operator,vert1_ref,vert1_volt,vert1_power'; 
                        END IF;
                    END IF;
                END IF;
                IF vert2_id_typ = 'n' THEN 
                    vert2_center = (SELECT ST_SetSRID(ST_MakePoint(lon/100.0,lat/100.0),900913) FROM planet_osm_nodes WHERE id = vert2_id);
                    vert2_name = (SELECT hstore(tags)->'name' FROM planet_osm_nodes WHERE id=vert2_id );
                    vert2_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_nodes WHERE id=vert2_id );
                    vert2_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_nodes WHERE id=vert2_id );
                    vert2_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_nodes WHERE id = vert2_id);
                    vert2_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_nodes WHERE id = vert2_id);
                    vert2_power = (SELECT value FROM vertices_ref_names JOIN planet_osm_nodes ON hstore(tags)->'power' = vertices_ref_names.name WHERE id = vert2_id);
                ELSE IF vert2_id_typ = 'w' THEN
                        vert2_center = (SELECT ST_centroid(way) FROM planet_osm_polygon WHERE osm_id = vert2_id);
                        vert2_name = (SELECT hstore(tags)->'name' FROM planet_osm_ways WHERE id=vert2_id );
                        vert2_ref= (SELECT hstore(tags)->'ref' FROM planet_osm_ways WHERE id=vert2_id );
                        vert2_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_ways WHERE id=vert2_id );
                        vert2_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_ways WHERE id = vert2_id);
                        vert2_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_ways WHERE id = vert2_id);
                        vert2_power = (SELECT value FROM vertices_ref_names JOIN planet_osm_ways ON  hstore(tags)->'power' = vertices_ref_names.name WHERE id = vert2_id);
                    ELSE IF vert2_id_typ = 'r' THEN
                        vert2_center = (SELECT ST_SetSRID(ST_MakePoint((max(lon) + min(lon))/200.0,(max(lat) + min(lat))/200.0),900913) FROM planet_osm_nodes WHERE id IN (
                                SELECT trim(leading 'n' from member)::bigint as node_parts FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = vert2_id) t WHERE member~E'[n]\\\\d+'
                                UNION
                                SELECT unnest(nodes) as node_parts FROM planet_osm_ways WHERE id IN (SELECT trim(leading 'w' from member)::bigint as way_parts FROM (SELECT unnest(members) as member FROM planet_osm_rels WHERE id = vert2_id) t WHERE member~E'[w]\\\\d+')));
                        vert2_name = (SELECT hstore(tags)->'name' FROM planet_osm_rels WHERE id=vert2_id );
                        vert2_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_rels WHERE id=vert2_id );
                        vert2_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_rels WHERE id=vert2_id );
                        vert2_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_rels WHERE id = vert2_id);
                        vert2_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_rels WHERE id = vert2_id);
                        vert2_power = (SELECT value FROM vertices_ref_names JOIN planet_osm_rels ON hstore(tags)->'power' = vertices_ref_names.name WHERE id = vert2_id);
                        ELSE 
                            INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_2subs','ERROR', 'Cannot get vert2_center,vert2_name,vert2_operator,vert2_ref,vert2_volt,vert2_power');
                            RETURN 'ERROR: Cannot get vert2_center,vert2_name,vert2_operator,vert2_ref,vert2_volt,vert2_power'; 
                        END IF;
                    END IF;
                END IF;
                link_length = (SELECT round(SUM(l)) FROM (SELECT ST_Length(ST_Transform(way,25832)) as l FROM planet_osm_line WHERE osm_id in (SELECT unnest(parts) FROM planet_osm_rels WHERE id = rel_id) and power in ('line','cable')) t2);
                link_geom = ST_MakeLine(vert1_center,vert2_center);
                link_volt = (SELECT (hstore(tags)->'voltage')::integer FROM planet_osm_rels WHERE id = rel_id);
                link_cables = (SELECT (hstore(tags)->'cables')::smallint FROM planet_osm_rels WHERE id = rel_id);
                link_wires = (SELECT hstore(tags)->'wires' FROM planet_osm_rels WHERE id = rel_id);
                link_wires_nb = (SELECT wires_nb FROM _analysis_rels WHERE osm_id = rel_id);
                link_frequency = (SELECT (hstore(tags)->'frequency')::text FROM planet_osm_rels WHERE id = rel_id);
                link_name = (SELECT (hstore(tags)->'name')::text FROM planet_osm_rels WHERE id = rel_id);
                link_operator = (SELECT (hstore(tags)->'operator')::text FROM planet_osm_rels WHERE id = rel_id);
                link_ref = (SELECT (hstore(tags)->'ref')::text FROM planet_osm_rels WHERE id = rel_id);
                link_transmissions = (SELECT array_agg(osm_id) FROM planet_osm_line WHERE osm_id in (SELECT unnest(parts) FROM planet_osm_rels WHERE id = rel_id) and power in ('line','cable'));

                INSERT INTO _vertices (osm_id,osm_id_typ,geo_center,longitude,latitude,role,voltage,from_relation) VALUES (vert1_id, vert1_id_typ, vert1_center, ST_X(ST_Transform(vert1_center,4326)), ST_Y(ST_Transform(vert1_center,4326)), vert1_power, vert1_volt, rel_id);
                INSERT INTO _vertices (osm_id,osm_id_typ,geo_center,longitude,latitude,role,voltage,from_relation) VALUES (vert2_id, vert2_id_typ, vert2_center, ST_X(ST_Transform(vert2_center,4326)), ST_Y(ST_Transform(vert2_center,4326)), vert2_power, vert2_volt, rel_id);
            
            -- Only insert as new vertice if it not exists in the final TABLE vertices
                IF not (SELECT exists(SELECT 1 FROM vertices_ref_id WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ)) THEN
                    INSERT INTO vertices_ref_id (osm_id,osm_id_typ,visible) VALUES(vert1_id,vert1_id_typ,'1'); 
                    link_id_1 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ);
                    INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(link_id_1,ST_X(ST_Transform(vert1_center,4326)),ST_Y(ST_Transform(vert1_center,4326)),vert1_power,vert1_volt,vert1_freq,vert1_name,vert1_operator,vert1_ref,ST_ASEWKT(ST_transform(vert1_center,4326)));
                ELSE
                    link_id_1 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ);
                    UPDATE vertices_ref_id SET visible = '1' WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ;
                    IF not (SELECT exists(SELECT 1 FROM vertices WHERE v_id=link_id_1)) THEN
                           INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(link_id_1,ST_X(ST_Transform(vert1_center,4326)),ST_Y(ST_Transform(vert1_center,4326)),vert1_power,vert1_volt,vert1_freq,vert1_name,vert1_operator,vert1_ref,ST_ASEWKT(ST_transform(vert1_center,4326))); 
                    END IF;
                END IF;
                IF not (SELECT exists(SELECT 1 FROM vertices_ref_id WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ)) THEN
                    INSERT INTO vertices_ref_id (osm_id,osm_id_typ,visible) VALUES(vert2_id,vert2_id_typ,'1');
                    link_id_2 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ);
                    INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(link_id_2,ST_X(ST_Transform(vert2_center,4326)),ST_Y(ST_Transform(vert2_center,4326)),vert2_power,vert2_volt,vert2_freq,vert2_name,vert2_operator,vert2_ref,ST_ASEWKT(ST_transform(vert2_center,4326)));
                ELSE 
                    link_id_2 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ);
                    UPDATE vertices_ref_id SET visible = '1' WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ;
                    IF not (SELECT exists(SELECT 1 FROM vertices WHERE v_id=link_id_2)) THEN
                           INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(link_id_2,ST_X(ST_Transform(vert2_center,4326)),ST_Y(ST_Transform(vert2_center,4326)),vert2_power,vert2_volt,vert2_freq,vert2_name,vert2_operator,vert2_ref,ST_ASEWKT(ST_transform(vert2_center,4326))); 
                    END IF;
                END IF;
             
                INSERT INTO _links (osm_id_1,osm_id_1_typ,osm_id_2,osm_id_2_typ,length_m,way,voltage,cables,wires,wires_nb,frequency,from_relation,from_transmissions) VALUES (vert1_id,vert1_id_typ, vert2_id,vert2_id_typ, link_length, link_geom, link_volt, link_cables, link_wires, link_wires_nb, link_frequency, rel_id, link_transmissions);
                INSERT INTO links (v_id_1,v_id_2,voltage,cables,wires,frequency,name,operator,ref,length_m,from_relation,WKT_SRID_4326) VALUES(link_id_1,link_id_2,link_volt,link_cables,link_wires_nb,link_frequency,link_name,link_operator,link_ref,link_length,rel_id,ST_ASEWKT(ST_transform(link_geom,4326)));
            
            UPDATE _analysis_rels SET abstracted = 1 WHERE osm_id = rel_id;
            
            INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_2subs','Done','Abstraction of relation with 2 substations: relation(' || rel_id || ');');
            RETURN 'Done. Abstraction of relation with 2 substations: relation(' || rel_id || ');';
            END;
            $$ LANGUAGE plpgsql;
            """
    cur.execute(sql)
    conn.commit()
    
    # Abstraction querry for relations with only 2 substations 
    sql =   """
            SELECT abstract_rel_with_2subs(rel_id) 
            FROM (
                SELECT osm_id as rel_id 
                FROM _analysis_rels 
                WHERE num_stations = 2 and incomplete = 'no') t
                ORDER BY rel_id;
            """            
    cur.execute(sql)
    msg = cur.fetchall()
    conn.commit()
    return msg
    

def get_w_nodes(way_id,cur,conn):
    # Obtain nodes from the extracted OSM "power" ways list.
    sql = "SELECT nodes FROM planet_osm_ways WHERE id = '" + str(way_id) + "';"
    cur.execute(sql)
    data = cur.fetchone()
    return data[0]

def ww_intersect(way1,way2,cur,conn):
    # Get intersection ways
    sql = "SELECT ST_Intersects((SELECT ST_Union(way) FROM planet_osm_line WHERE osm_id = " + str(way1) + "),(SELECT ST_Union(way) FROM planet_osm_line WHERE osm_id = " + str(way2) + "));"
    cur.execute(sql)
    data = cur.fetchone()
    return data[0]

def wsub_intersect(way,substation,cur,conn):
    # Returns true if way and substation intersect otherwise false.
    sql = "SELECT ST_Intersects((SELECT ST_Union(way) FROM planet_osm_line WHERE osm_id = " + str(way) + "),(SELECT way FROM planet_osm_polygon WHERE osm_id = " + str(substation) + "));"
    cur.execute(sql)
    data = cur.fetchone()
    return data[0]

def separate_parts(station_ids, transmission_ids, T_node_id,cur,conn):
    """ Extend each section, beginning at T-junction and going to substations"""
    l1,l2,l3,rest = [],[],[],[]
    # Separate three ways that start at T-junction from transmission sections
    for t_id in transmission_ids:
        if T_node_id in get_w_nodes(t_id,cur,conn):
            if len(l1) == 0:
                l1.append(t_id)
            elif len(l2) == 0:
                l2.append(t_id)
            else:
                l3.append(t_id)
        else:
            rest.append(t_id)  
    
    if len(l1) + len(l2) + len(l3) != 3:
        return "ERROR: Did not find three transmission lines starting at T-junction."
    
    iteration = 0
    while len(rest) > 0 and iteration < 100:
        if ww_intersect(rest[-1],l1[-1],cur,conn):
            l1.append(rest.pop())
        elif ww_intersect(rest[-1],l2[-1],cur,conn):
            l2.append(rest.pop())
        elif ww_intersect(rest[-1],l3[-1],cur,conn):
            l3.append(rest.pop())
        else:
            # move last entry of rest to first entry
            last = rest.pop()
            rest = [last] + rest
        iteration += 1
        
    if len(rest) > 0:
        return "ERROR: Could not connect transmission segments: %s" %(str(rest).translate(None,"'[]L "))
    else:
        TBA_station = False
        ordered_stations = [0,0,0]
        for station in station_ids:
            if wsub_intersect(l1[-1],station,cur,conn) and ordered_stations[0] == 0:
                ordered_stations[0] = station
            elif wsub_intersect(l2[-1],station,cur,conn) and ordered_stations[1] == 0:
                ordered_stations[1] = station
            elif wsub_intersect(l3[-1],station,cur,conn) and ordered_stations[2] == 0:
                ordered_stations[2] = station
            elif not TBA_station:
                TBA_station = station
            else:
                return "ERROR: Could not abstract this relation, too many unconnected segments to a substation."
        if TBA_station != False:
            ordered_stations[ordered_stations.index(0)] = TBA_station
    T_node_id
    return (l1,l2,l3,ordered_stations)

def insert_segments(rel_id,T_node_id,stations,result,cur,conn):
    l1_ids = str(result[0]).translate(None,"'[]L ")
    l2_ids = str(result[1]).translate(None,"'[]L ")  
    l3_ids = str(result[2]).translate(None,"'[]L ")
    ordered_station_ids = result[3]
    ordered_station_types = []
    ordered_station_roles = []
    for o_stat in ordered_station_ids:
        for sta in stations:
            if str(o_stat) in str(sta):
                ordered_station_types.append(sta[0])
                # TODO: Clear this section. Vertices names are replaced by "substation" if they have the name sub_station, station  
                role = re.sub(r"\d", "", sta[1:])
                if role in ['sub_station','station']:
                    role = 'substation'               
                ordered_station_roles.append(role)
    sql =   """
            SELECT * FROM abstract_rel_with_T_node(
            '%s', '%s', '{%s}'::bigint[], '{%s}'::bigint[], '{%s}'::bigint[],
            '{%s}'::bigint[], '{%s}'::text[], '{%s}'::text[]);
            """ %(str(rel_id),str(T_node_id),l1_ids,l2_ids,l3_ids, \
                  str(ordered_station_ids).translate(None,"'[]L "), \
                  str(ordered_station_types).translate(None,"'[]L "), \
                  str(ordered_station_roles).translate(None,"'[]L "))
    cur.execute(sql)
    msg = cur.fetchone()
    conn.commit()
    return msg[0]

# Query to abstract T-junctions    
def abstract_3subs_T_node(cur,conn):   
    sql =   """
            CREATE OR REPLACE FUNCTION abstract_rel_with_T_node(
            rel_id                  bigint, 
            T_node_id               bigint,  
            link1_transmissions     bigint[],
            link2_transmissions     bigint[],
            link3_transmissions     bigint[],
            ordered_station_ids     bigint[],
            ordered_station_types   text[],
            ordered_station_roles   text[])
            RETURNS text
                            AS $$
                        DECLARE  T_node_loc         geometry;
                                 vert1_id           bigint;
                                 vert2_id           bigint;
                                 vert3_id           bigint;
                                 vert1_id_typ       char;
                                 vert2_id_typ       char;
                                 vert3_id_typ       char;
                                 vert1_center       geometry;
                                 vert2_center       geometry;
                                 vert3_center       geometry;
                                 vert1_power        text;
                                 vert2_power        text;
                                 vert3_power        text;
                                 vert1_volt         text;
                                 vert2_volt         text;
                                 vert3_volt         text;
                                 vert1_freq         text;
                                 vert2_freq         text;
                                 vert3_freq         text;
                                 vert1_name         text;
                                 vert2_name         text;
                                 vert3_name         text;
                                 vert1_operator     text;
                                 vert2_operator     text;
                                 vert3_operator     text;
                                 vert1_ref          text;
                                 vert2_ref          text;
                                 vert3_ref          text;
                                 link1_len          integer;
                                 link2_len          integer;
                                 link3_len          integer;
                                 link1_way          geometry;
                                 link2_way          geometry;
                                 link3_way          geometry;
                                 rel_volt           integer;
                                 rel_cables         smallint; 
                                 rel_wires          text; 
                                 rel_wires_nb       smallint;
                                 rel_frequency      text; 
                                 rel_operator       text;
                                 rel_name           text;
                                 rel_ref            text;
                                 id_1               bigint;
                                 id_2               bigint; 
                                 id_3               bigint;
                                 id_4               bigint;
            BEGIN           
            T_node_loc = (SELECT ST_SetSRID(ST_MakePoint(lon/100.0,lat/100.0),900913) FROM planet_osm_nodes WHERE id = T_node_id);
            vert1_id = ordered_station_ids[1]; 
            vert2_id = ordered_station_ids[2];
            vert3_id = ordered_station_ids[3];
            vert1_id_typ = ordered_station_types[1];
            vert2_id_typ = ordered_station_types[2];
            vert3_id_typ = ordered_station_types[3];
            vert1_power = ordered_station_roles[1];
            vert2_power = ordered_station_roles[2];
            vert3_power = ordered_station_roles[3];
            
            IF vert1_id_typ = 'n' THEN 
                vert1_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_nodes WHERE id = vert1_id);
                vert1_center = (SELECT ST_SetSRID(ST_MakePoint(lon/100.0,lat/100.0),900913) FROM planet_osm_nodes WHERE id = vert1_id);
                vert1_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_nodes WHERE id = vert1_id);
                vert1_name = (SELECT hstore(tags)->'name' FROM planet_osm_nodes WHERE id = vert1_id);
                vert1_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_nodes WHERE id = vert1_id);
                vert1_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_nodes WHERE id = vert1_id);
            ELSE IF vert1_id_typ = 'w' THEN
                vert1_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_ways WHERE id = vert1_id);
                vert1_center = (SELECT ST_centroid(way) FROM planet_osm_polygon WHERE osm_id = vert1_id);
                vert1_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_ways WHERE id = vert1_id);
                vert1_name = (SELECT hstore(tags)->'name' FROM planet_osm_ways WHERE id = vert1_id);
                vert1_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_ways WHERE id = vert1_id);
                vert1_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_ways WHERE id = vert1_id);
            ELSE IF vert1_id_typ = 'r' THEN
                vert1_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_rels WHERE id = vert1_id);
                vert1_center = (SELECT ST_SetSRID(ST_MakePoint((max(lon) + min(lon))/200.0,(max(lat) + min(lat))/200.0),900913) 
                    FROM planet_osm_nodes
                    WHERE id IN (
                        SELECT trim(leading 'n' from member)::bigint as node_parts FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = rel_id) t WHERE member~E'[n]\\d+'
                        UNION
                        SELECT unnest(nodes) as node_parts FROM planet_osm_ways WHERE id IN (SELECT trim(leading 'w' from member)::bigint as way_parts 
                        FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = rel_id) t WHERE member~E'[w]\\d+')));
                vert1_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_rels WHERE id = vert1_id);
                vert1_name = (SELECT hstore(tags)->'name' FROM planet_osm_rels WHERE id = vert1_id);
                vert1_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_rels WHERE id = vert1_id);
                vert1_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_rels WHERE id = vert1_id);
            ELSE 
                INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_T_node','ERROR','Cannot get vert1_center: relation(' || rel_id || ');');
                RETURN 'ERROR: Cannot get vert1_center';
            END IF;
            END IF;
            END IF;
            
            IF vert2_id_typ = 'n' THEN 
                vert2_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_nodes WHERE id = vert2_id);
                vert2_center = (SELECT ST_SetSRID(ST_MakePoint(lon/100.0,lat/100.0),900913) FROM planet_osm_nodes WHERE id = vert2_id);
                vert2_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_nodes WHERE id = vert2_id);
                vert2_name = (SELECT hstore(tags)->'name' FROM planet_osm_nodes WHERE id = vert2_id);
                vert2_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_nodes WHERE id = vert2_id);
                vert2_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_nodes WHERE id = vert2_id);
            ELSE IF vert2_id_typ = 'w' THEN
                vert2_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_ways WHERE id = vert2_id);
                vert2_center = (SELECT ST_centroid(way) FROM planet_osm_polygon WHERE osm_id = vert2_id);
                vert2_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_ways WHERE id = vert2_id);
                vert2_name = (SELECT hstore(tags)->'name' FROM planet_osm_ways WHERE id = vert2_id);
                vert2_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_ways WHERE id = vert2_id);
                vert2_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_ways WHERE id = vert2_id);
            ELSE IF vert2_id_typ = 'r' THEN
                vert2_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_rels WHERE id = vert2_id);
                vert2_center = (SELECT ST_SetSRID(ST_MakePoint((max(lon) + min(lon))/200.0,(max(lat) + min(lat))/200.0),900913) FROM planet_osm_nodes WHERE id IN (
                            SELECT trim(leading 'n' from member)::bigint as node_parts FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = rel_id) t WHERE member~E'[n]\\d+'
                            UNION
                            SELECT unnest(nodes) as node_parts FROM planet_osm_ways WHERE id IN (SELECT trim(leading 'w' from member)::bigint as way_parts FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = rel_id) t WHERE member~E'[w]\\d+')));
                vert2_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_rels WHERE id = vert2_id);
                vert2_name = (SELECT hstore(tags)->'name' FROM planet_osm_rels WHERE id = vert2_id);
                vert2_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_rels WHERE id = vert2_id);
                vert2_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_rels WHERE id = vert2_id);
            ELSE 
                INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_T_node','ERROR','Cannot get vert2_center: relation(' || rel_id || ');');
                RETURN 'ERROR: Cannot get vert2_center';
            END IF;
            END IF;
            END IF;

            IF vert3_id_typ = 'n' THEN 
                vert3_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_nodes WHERE id = vert3_id);
                vert3_center = (SELECT ST_SetSRID(ST_MakePoint(lon/100.0,lat/100.0),900913) FROM planet_osm_nodes WHERE id = vert3_id);
                vert3_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_nodes WHERE id = vert3_id);
                vert3_name = (SELECT hstore(tags)->'name' FROM planet_osm_nodes WHERE id = vert3_id);
                vert3_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_nodes WHERE id = vert3_id);
                vert3_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_nodes WHERE id = vert3_id);
            ELSE IF vert3_id_typ = 'w' THEN
                vert3_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_ways WHERE id = vert3_id);
                vert3_center = (SELECT ST_centroid(way) FROM planet_osm_polygon WHERE osm_id = vert3_id);
                vert3_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_ways WHERE id = vert3_id);
                vert3_name = (SELECT hstore(tags)->'name' FROM planet_osm_ways WHERE id = vert3_id);
                vert3_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_ways WHERE id = vert3_id);
                vert3_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_ways WHERE id = vert3_id);
            ELSE IF vert3_id_typ = 'r' THEN
                vert3_volt = (SELECT hstore(tags)->'voltage' FROM planet_osm_rels WHERE id = vert3_id);
                vert3_center = (SELECT ST_SetSRID(ST_MakePoint((max(lon) + min(lon))/200.0,(max(lat) + min(lat))/200.0),900913) FROM planet_osm_nodes WHERE id IN (
                        SELECT trim(leading 'n' from member)::bigint as node_parts FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = vert3_id) t WHERE member~E'[n]\\d+'
                        UNION
                        SELECT unnest(nodes) as node_parts FROM planet_osm_ways WHERE id IN (SELECT trim(leading 'w' from member)::bigint as way_parts FROM (SELECT unnest(members) as member,* FROM planet_osm_rels WHERE id = vert3_id) t WHERE member~E'[w]\\d+')));
                vert3_freq = (SELECT hstore(tags)->'frequency' FROM planet_osm_rels WHERE id = vert3_id);
                vert3_name = (SELECT hstore(tags)->'name' FROM planet_osm_rels WHERE id = vert3_id);
                vert3_operator = (SELECT hstore(tags)->'operator' FROM planet_osm_rels WHERE id = vert3_id);
                vert3_ref = (SELECT hstore(tags)->'ref' FROM planet_osm_rels WHERE id = vert3_id);
            ELSE 
                INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_T_node','ERROR','Cannot get vert3_center: relation(' || rel_id || ');');
                RETURN 'ERROR: Cannot get vert3_center'; 
            END IF;
            END IF;
            END IF;
            
            link1_way = ST_MakeLine(vert1_center, T_node_loc);
            link2_way = ST_MakeLine(vert2_center, T_node_loc);
            link3_way = ST_MakeLine(vert3_center, T_node_loc);    
            link1_len = (SELECT round(SUM(l)) FROM (SELECT ST_Length(ST_Transform(way,25832)) as l FROM planet_osm_line WHERE osm_id = ANY (link1_transmissions) AND power in ('line','cable'))t); 
            link2_len = (SELECT round(SUM(l)) FROM (SELECT ST_Length(ST_Transform(way,25832)) as l FROM planet_osm_line WHERE osm_id = ANY (link2_transmissions) AND power in ('line','cable'))t); 
            link3_len = (SELECT round(SUM(l)) FROM (SELECT ST_Length(ST_Transform(way,25832)) as l FROM planet_osm_line WHERE osm_id = ANY (link3_transmissions) AND power in ('line','cable'))t); 
            rel_volt = (SELECT (hstore(tags)->'voltage')::text FROM planet_osm_rels WHERE id = rel_id);
            rel_cables = (SELECT (hstore(tags)->'cables')::text FROM planet_osm_rels WHERE id = rel_id);
            rel_wires = (SELECT (hstore(tags)->'wires')::text FROM planet_osm_rels WHERE id = rel_id);
            rel_wires_nb = (SELECT wires_nb FROM _analysis_rels WHERE osm_id = rel_id);
            rel_frequency = (SELECT (hstore(tags)->'frequency')::text FROM planet_osm_rels WHERE id = rel_id);
            rel_name = (SELECT (hstore(tags)->'name')::text FROM planet_osm_rels WHERE id = rel_id);
            rel_operator = (SELECT (hstore(tags)->'operator')::text FROM planet_osm_rels WHERE id = rel_id);
            rel_ref = (SELECT (hstore(tags)->'ref')::text FROM planet_osm_rels WHERE id = rel_id);
                            
            INSERT INTO _vertices (osm_id,osm_id_typ,geo_center,longitude,latitude,role,voltage,from_relation)  VALUES(vert1_id, vert1_id_typ, vert1_center, ST_X(ST_Transform(vert1_center,4326)), ST_Y(ST_Transform(vert1_center,4326)), vert1_power, rel_volt, rel_id);
            INSERT INTO _vertices (osm_id,osm_id_typ,geo_center,longitude,latitude,role,voltage,from_relation)  VALUES(vert2_id, vert2_id_typ, vert2_center, ST_X(ST_Transform(vert2_center,4326)), ST_Y(ST_Transform(vert2_center,4326)), vert2_power, rel_volt, rel_id);
            INSERT INTO _vertices (osm_id,osm_id_typ,geo_center,longitude,latitude,role,voltage,from_relation)  VALUES(vert3_id, vert3_id_typ, vert3_center, ST_X(ST_Transform(vert3_center,4326)), ST_Y(ST_Transform(vert3_center,4326)), vert3_power, rel_volt, rel_id);
            INSERT INTO _vertices (osm_id,osm_id_typ,geo_center,longitude,latitude,role,voltage,from_relation)  VALUES(T_node_id, 'a', T_node_loc, ST_X(ST_Transform(T_node_loc,4326)), ST_Y(ST_Transform(T_node_loc,4326)), 'auxillary_T_node', rel_volt, rel_id);
            
            
            -- Only insert as new vertice if it not exists in the final TABLE vertices
            IF not (SELECT exists(SELECT 1 FROM vertices_ref_id WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ)) THEN
                    INSERT INTO vertices_ref_id (osm_id,osm_id_typ,visible) VALUES(vert1_id,vert1_id_typ,'1'); 
                    id_1 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ);
                    INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_1,ST_X(ST_Transform(vert1_center,4326)),ST_Y(ST_Transform(vert1_center,4326)),vert1_power,vert1_volt,vert1_freq,vert1_name,vert1_operator,vert1_ref,ST_ASEWKT(ST_transform(vert1_center,4326)));
            ELSE
                    id_1 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ);
                    UPDATE vertices_ref_id SET visible = '1' WHERE osm_id=vert1_id and osm_id_typ=vert1_id_typ;
                    IF not (SELECT exists(SELECT 1 FROM vertices WHERE v_id=id_1)) THEN
                            INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_1,ST_X(ST_Transform(vert1_center,4326)),ST_Y(ST_Transform(vert1_center,4326)),vert1_power,vert1_volt,vert1_freq,vert1_name,vert1_operator,vert1_ref,ST_ASEWKT(ST_transform(vert1_center,4326))); 
                    END IF;
            END IF;
            IF not (SELECT exists(SELECT 1 FROM vertices_ref_id WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ)) THEN
                    INSERT INTO vertices_ref_id (osm_id,osm_id_typ,visible) VALUES(vert2_id,vert2_id_typ,'1');
                    id_2 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ);
                    INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_2,ST_X(ST_Transform(vert2_center,4326)),ST_Y(ST_Transform(vert2_center,4326)),vert2_power,vert2_volt,vert2_freq,vert2_name,vert2_operator,vert2_ref,ST_ASEWKT(ST_transform(vert2_center,4326)));
            ELSE 
                    id_2 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ);
                    UPDATE vertices_ref_id SET visible = '1' WHERE osm_id=vert2_id and osm_id_typ=vert2_id_typ;
                    IF not (SELECT exists(SELECT 1 FROM vertices WHERE v_id=id_2)) THEN
                            INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_2,ST_X(ST_Transform(vert2_center,4326)),ST_Y(ST_Transform(vert2_center,4326)),vert2_power,vert2_volt,vert2_freq,vert2_name,vert2_operator,vert2_ref,ST_ASEWKT(ST_transform(vert2_center,4326))); 
                    END IF;
            END IF;
            IF not (SELECT exists(SELECT 1 FROM vertices_ref_id WHERE osm_id=vert3_id and osm_id_typ=vert3_id_typ)) THEN
                    INSERT INTO vertices_ref_id (osm_id,osm_id_typ,visible) VALUES(vert3_id,vert3_id_typ,'1');
                    id_3 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert3_id and osm_id_typ=vert3_id_typ);
                    INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_3,ST_X(ST_Transform(vert3_center,4326)),ST_Y(ST_Transform(vert3_center,4326)),vert3_power,vert3_volt,vert3_freq,vert3_name,vert3_operator,vert3_ref,ST_ASEWKT(ST_transform(vert3_center,4326)));
            ELSE 
                    id_3 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=vert3_id and osm_id_typ=vert3_id_typ);
                    UPDATE vertices_ref_id SET visible = '1' WHERE osm_id=vert3_id and osm_id_typ=vert3_id_typ;
                    IF not (SELECT exists(SELECT 1 FROM vertices WHERE v_id=id_3)) THEN
                            INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_3,ST_X(ST_Transform(vert3_center,4326)),ST_Y(ST_Transform(vert3_center,4326)),vert3_power,vert3_volt,vert3_freq,vert3_name,vert3_operator,vert3_ref,ST_ASEWKT(ST_transform(vert3_center,4326))); 
                    END IF;
            END IF;
            IF not (SELECT exists(SELECT 1 FROM vertices_ref_id WHERE osm_id=T_node_id and osm_id_typ='a')) THEN
                    INSERT INTO vertices_ref_id (osm_id,osm_id_typ,visible) VALUES(T_node_id,'a','1');
                    id_4 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=T_node_id and osm_id_typ='a');
                    INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_4,ST_X(ST_Transform(T_node_loc,4326)),ST_Y(ST_Transform(T_node_loc,4326)),'auxillary_T_node',rel_volt,rel_frequency,rel_name,rel_operator,rel_ref,ST_ASEWKT(ST_transform(T_node_loc,4326)));
            ELSE 
                    id_4 = (SELECT v_id FROM vertices_ref_id WHERE osm_id=T_node_id and osm_id_typ='a');
                    UPDATE vertices_ref_id SET visible = '1' WHERE osm_id=T_node_id and osm_id_typ='a';
                    IF not (SELECT exists(SELECT 1 FROM vertices WHERE v_id=id_4)) THEN
                            INSERT INTO vertices (v_id,lon,lat,typ,voltage,frequency,name,operator,ref,WKT_SRID_4326) VALUES(id_4,ST_X(ST_Transform(T_node_loc,4326)),ST_Y(ST_Transform(T_node_loc,4326)),'auxillary_T_node',rel_volt,rel_frequency,rel_name,rel_operator,rel_ref,ST_ASEWKT(ST_transform(T_node_loc,4326))); 
                    END IF;
            END IF;
            
            INSERT INTO _links (osm_id_1,osm_id_1_typ,osm_id_2,osm_id_2_typ,length_m,way,voltage,cables,wires,wires_nb,frequency,from_relation,from_transmissions) VALUES (vert1_id,vert1_id_typ, T_node_id,'a', link1_len, link1_way, rel_volt, rel_cables, rel_wires, rel_wires_nb, rel_frequency, rel_id, link1_transmissions);
            INSERT INTO _links (osm_id_1,osm_id_1_typ,osm_id_2,osm_id_2_typ,length_m,way,voltage,cables,wires,wires_nb,frequency,from_relation,from_transmissions) VALUES (vert2_id,vert2_id_typ, T_node_id,'a', link2_len, link2_way, rel_volt, rel_cables, rel_wires, rel_wires_nb, rel_frequency, rel_id, link2_transmissions);
            INSERT INTO _links (osm_id_1,osm_id_1_typ,osm_id_2,osm_id_2_typ,length_m,way,voltage,cables,wires,wires_nb,frequency,from_relation,from_transmissions) VALUES (vert3_id,vert3_id_typ, T_node_id,'a', link3_len, link3_way, rel_volt, rel_cables, rel_wires, rel_wires_nb, rel_frequency, rel_id, link3_transmissions);
            INSERT INTO links (v_id_1,v_id_2,voltage,cables,wires,frequency,name,operator,ref,length_m,from_relation,WKT_SRID_4326) VALUES(id_1,id_4,rel_volt, rel_cables, rel_wires_nb, rel_frequency, rel_name, rel_operator, rel_ref, link1_len,rel_id,ST_ASEWKT(ST_transform(link1_way,4326)));
            INSERT INTO links (v_id_1,v_id_2,voltage,cables,wires,frequency,name,operator,ref,length_m,from_relation,WKT_SRID_4326) VALUES(id_2,id_4,rel_volt, rel_cables, rel_wires_nb, rel_frequency, rel_name, rel_operator, rel_ref, link2_len,rel_id,ST_ASEWKT(ST_transform(link2_way,4326)));
            INSERT INTO links (v_id_1,v_id_2,voltage,cables,wires,frequency,name,operator,ref,length_m,from_relation,WKT_SRID_4326) VALUES(id_3,id_4,rel_volt, rel_cables, rel_wires_nb, rel_frequency, rel_name, rel_operator, rel_ref, link3_len,rel_id,ST_ASEWKT(ST_transform(link3_way,4326)));
                        
                        UPDATE _analysis_rels SET abstracted = 1, t_node_ids = array[T_node_id]::bigint[] WHERE osm_id = rel_id;
                        
                        INSERT INTO msg_log (function,typ,message) VALUES ('abstract_rel_with_T_node','Done','Abstraction of relation with 3 stations and T-node: relation(' || rel_id || ');');
                        RETURN 'Done. Abstraction of relation with 3 stations and T-node: relation(' || rel_id || ');';
                            END;
                            $$ LANGUAGE plpgsql;
            """    
    cur.execute(sql)
    conn.commit()
    
    # Get all T-junction relations with 3 stations
    sql =   """
            SELECT osm_id,t_node_ids,stations,transmissions
            FROM _analysis_rels 
            WHERE t_node_ids <> '{}' AND incomplete = 'no' AND num_stations = 3 
            ORDER BY osm_id;
            """
    cur.execute(sql)
    result = cur.fetchall()
    success = 0
    for (rel_id,T_node_ids,stations,transmissions) in result:
        #print rel_id
        station_ids = [re.sub(r"\D", "", s) for s in stations] 
        station_ids_typ = [s[0] for s in stations]
        transmission_ids = [re.sub(r"\D", "", t) for t in transmissions] 
        
        # FIXME: stations from relation type have no geometry, 
        # How to find a intersection of a line segment with the relation if the relation is a power plant?
        if 'r' in station_ids_typ:
            continue

        for T_node_id in T_node_ids:
            res = separate_parts(station_ids, transmission_ids, T_node_id,cur,conn)                 
            if type(res) is tuple:              
                break

        if type(res) is str:
            logging.warning(res)
            sql =   """
                    INSERT INTO _t_node_problems (rel_id,t_node_ids,stations,transmissions,error_msg) 
                    VALUES('%s','{%s}'::bigint[],'{%s}'::text[],'{%s}'::text[],'%s');
                    """  % (str(rel_id),str(T_node_ids).translate(None,"'[]L "),str(stations).translate(None,"'[]L "),str(transmissions).translate(None,"'[]L "),res)
            cur.execute(sql)
            conn.commit()
            continue
        else:
            msg = insert_segments(rel_id,T_node_id,stations,res,cur,conn)
            success += 1
    
    return success

# TODO:
def abstract_3subs_in_row(cur,conn):
    pass

def rel_abstraction(cur,conn):
    msg = abstract_2subs(cur,conn)
    success = abstract_3subs_T_node(cur,conn)
    #abstract_3subs_in_row(cur,conn)
    return len(msg) + success
    
if __name__ == '__main__':
    try:
        conn       = dbconn_from_args()
        cur        = conn.cursor()
        nr_success = rel_abstraction(cur,conn)
        logging.info('Number of abstracted relations: ' +  str(nr_success))
    except Exception, e:
        logging.error('Could not abstract relations in database.', exc_info=True)
        exit(1)
    
